using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.IO;
using System.Collections;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Printing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using Syncfusion.Runtime.InteropServices;

namespace ZoomRichText
{
	public enum BackgroundMode
	{
		TRANSPARENT = 1,
		OPAQUE = 2
	}


	/// <summary>
	/// Summary description for RichText2EMF.
	/// </summary>
	public class RichText2EMF
	{
		#region RTF to WMF stuff

		[ StructLayout( LayoutKind.Sequential )]
			private struct RECT 
		{
			public int left;
			public int top;
			public int right;
			public int bottom;
		}
		[ StructLayout( LayoutKind.Sequential )]
			private struct CHARRANGE 
		{
			public int cpMin;
			public int cpMax;
		}
		[ StructLayout( LayoutKind.Sequential )]
			private struct FORMATRANGE 
		{
			public IntPtr hdc; 
			public IntPtr hdcTarget; 
			public RECT rc; 
			public RECT rcPage; 
			public CHARRANGE chrg; 
		}

		[DllImport("user32.dll", CharSet=CharSet.Auto)]
		private static extern IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam);
		[DllImport("gdi32.dll", CharSet=CharSet.Auto)]
		public static extern int SetBkMode(IntPtr hDC, BackgroundMode mode);

		private const int WM_USER			 = 0x400;
		private const int EM_FORMATRANGE	 = WM_USER+57;
		private const int EM_SETTARGETDEVICE = WM_USER+72;

		#endregion

		static private EmfType EMF_TYPE = EmfType.EmfPlusOnly;

		/// <summary>
		/// Convert between 1/100 inch (unit used by the .NET framework)
		/// and twips (1/1440 inch, used by Win32 API calls)
		/// </summary>
		/// <param name="n">Value in 1/100 inch</param>
		/// <returns>Value in twips</returns>
		static private int HundredthInchToTwips(int n)
		{
			return (int)((float)n*14.4f*1.04f);
		}

		static private int HundredthInchToTwips(float f)
		{
			return (int)(f*14.4f*1.04f);
		}

		public static byte[] BytesFromRichText(float width, float height, string text)
		{
			byte[] img;

			RichTextBox richTextBox = new RichTextBox();
			richTextBox.AcceptsTab = true;
			richTextBox.BorderStyle = System.Windows.Forms.BorderStyle.None;

			richTextBox.Bounds = new System.Drawing.Rectangle(0, 0, (int)width, (int)height);

			if(text.StartsWith(@"{\rtf"))
			{
				richTextBox.Rtf = text;
			}
			else
			{
				richTextBox.Text = text;
			}

			float top = 0;

			// fix due to low data quality of source db
			// check if height or width are zero and increase to 1 so that
			// following calls won't fail
			height = (height == 0) ? 1 : height;
			width = (width == 0) ? 1 : width;

			// Create a Bitmap
			Bitmap bitmap = new Bitmap(1, 1, PixelFormat.Format24bppRgb);
			// Wrap a Graphics around the Bitmap
			Graphics rtfGraphics = Graphics.FromImage( bitmap );
			IntPtr refHdc = rtfGraphics.GetHdc();

			// paint RichTextBox content to MetaFile
			img = BytesFromRichText(richTextBox, refHdc, 0, top, width, height);
			
			// and remove non-transparent background
			try
			{
				img = RemoveBackground(img, refHdc, width, height);
			}
			catch(Exception)
			{
			}

			rtfGraphics.ReleaseHdc(refHdc);
			rtfGraphics.Dispose();
			bitmap.Dispose();
			richTextBox.Dispose();

			return img;
		}

		private static byte[] BytesFromRichText(RichTextBox richTextBox, IntPtr refHdc, float left, float top, float width, float height)
		{
			MemoryStream ms = new MemoryStream();
			Metafile mf = new Metafile(ms, refHdc, new RectangleF(0, 0, width, height), MetafileFrameUnit.Pixel, EMF_TYPE);

			Graphics g = Graphics.FromImage(mf);
			g.TextRenderingHint = TextRenderingHint.AntiAliasGridFit;
			IntPtr hdc = g.GetHdc();
			
			CHARRANGE cr;
			cr.cpMin = 0;
			cr.cpMax = richTextBox.TextLength;

			RECT rc;	// margins
			rc.left		= HundredthInchToTwips(left);
			rc.top		= HundredthInchToTwips(top);
			rc.right	= HundredthInchToTwips(width);
			rc.bottom	= HundredthInchToTwips(height);

			RECT rcPage;	// page bounds
			rcPage.left		= 0;
			rcPage.top		= 0;
			rcPage.right	= 0;
			rcPage.bottom	= 0;

			FORMATRANGE fr;
			fr.chrg		 = cr;
			fr.hdc		 = hdc;
			fr.hdcTarget = hdc;
			fr.rc		 = rc;
			fr.rcPage	 = rc; //Page;

			IntPtr res;

			IntPtr wpar = new IntPtr(1);
			IntPtr lpar = Marshal.AllocCoTaskMem( Marshal.SizeOf( fr ) ); 
			Marshal.StructureToPtr(fr, lpar, false);

			res = SendMessage(richTextBox.Handle, EM_FORMATRANGE, wpar, lpar);

			// cleanup
			Marshal.FreeCoTaskMem(lpar);

			wpar = new IntPtr(0);
			lpar = new IntPtr(0);
			res = SendMessage(richTextBox.Handle, EM_FORMATRANGE, wpar, lpar);

			g.ReleaseHdc(hdc);
			g.Dispose();

			ms.Close();
			mf.Dispose();

			return ms.ToArray();
		}

		private static byte[] RemoveBackground(byte[] metafile, IntPtr refHdc, float width, float height)
		{
			Metafile mf1 = new Metafile(new MemoryStream(metafile));
			MemoryStream ms = new MemoryStream();
			Metafile mf = new Metafile(ms, refHdc, new RectangleF(0, 0, width, height), MetafileFrameUnit.Pixel, EMF_TYPE);

			Graphics g = Graphics.FromImage(mf);
			g.TextRenderingHint = TextRenderingHint.AntiAliasGridFit;

			g.TranslateTransform(1, 2);
			IntPtr hdc = g.GetHdc();

			SetBkMode(hdc, BackgroundMode.TRANSPARENT);

			g.ReleaseHdc(hdc);

			MetaFileCallbackFilter filter = new MetaFileCallbackFilter(mf1);
			Graphics.EnumerateMetafileProc fnCallback = new	Graphics.EnumerateMetafileProc(filter.CallbackFilter );
			g.EnumerateMetafile( mf1, new Point( 0, 0 ), fnCallback, IntPtr.Zero );

			g.Dispose();

			ms.Close();
			mf.Dispose();
			mf1.Dispose();

			return ms.ToArray();
		}
	}


	public class MetaFileCallbackFilter
	{
		private Metafile mf = null;
		
		public MetaFileCallbackFilter(Metafile mf)
		{
			this.mf = mf;
		}

		public bool CallbackFilter(
			EmfPlusRecordType recordType,
			int flags,
			int dataSize,
			IntPtr data,
			PlayRecordCallback callbackData)
		{
			byte[] dataArray;
			if(dataSize > 0)
			{
				dataArray = new byte[dataSize];
				Marshal.Copy(data, dataArray, 0, dataSize);
			}
			else dataArray = new byte[0];

			switch(recordType)
			{
				case EmfPlusRecordType.EmfExtTextOutW:
					// clear fOptions
					dataArray[14*4+0] = 0;
					dataArray[14*4+1] = 0;
					dataArray[14*4+2] = 0;
					dataArray[14*4+3] = 0;
					this.mf.PlayRecord(recordType, flags, dataSize, dataArray);
					break;
				case EmfPlusRecordType.EmfSetBkColor:
					// remove
					break;
				case EmfPlusRecordType.EmfSetBkMode:
					dataArray[0] = (byte)BackgroundMode.TRANSPARENT;
					this.mf.PlayRecord(recordType, flags, dataSize, dataArray);
					break;
				default:
					this.mf.PlayRecord(recordType, flags, dataSize, dataArray);
					break;
			}	
			dataArray = null;
			return true;
		}
	}
}
